from vclibpy.components.compressors.compressor import Compressor
from vclibpy.datamodels import Inputs


class RotaryCompressor(Compressor):
    """
    Compressor model based on the thesis of Mirko Engelpracht.

    This compressor is characterized by using regressions provided by Mirko Engelpracht for a family of rotary
    compressors. The coefficients used in the regressions are sourced from his Master's thesis.

    Parameters:
        N_max (float): Maximal rotations per second of the compressor.
        V_h (float): Volume of the compressor in m^3.

    Methods:
        get_lambda_h(inputs: Inputs) -> float:
            Returns the volumetric efficiency based on the regressions of Mirko Engelpracht.

        get_eta_isentropic(p_outlet: float, inputs: Inputs) -> float:
            Returns the isentropic efficiency based on the regressions of Mirko Engelpracht.

        get_eta_mech(inputs: Inputs) -> float:
            Returns the mechanical efficiency based on the regressions of Mirko Engelpracht.

    """

    def get_lambda_h(self, inputs: Inputs) -> float:
        """
        Returns the volumetric efficiency based on the regressions of Mirko Engelpracht.

        Args:
            inputs (Inputs): Input parameters for the calculation.

        Returns:
            float: Volumetric efficiency.
        """
        p_outlet = self.get_p_outlet()
        # If not constant value is given, eta_is is calculated based on the regression of Mirko Engelpracht
        n = self.get_n_absolute(inputs.n)
        T_1 = self.state_inlet.T

        a_1 = 0.80179
        a_2 = -0.05210
        sigma_pi = 1.63495
        pi_ave = 4.54069
        a_3 = 3.21616e-4
        sigma_T_1 = 8.43797
        T_1_ave = 263.86428
        a_4 = -0.00494
        a_5 = 0.04981
        sigma_n = 20.81378
        n_ave = 64.41071
        a_6 = -0.02190

        pi = p_outlet / self.state_inlet.p
        return (
                a_1 +
                a_2 * (pi - pi_ave) / sigma_pi +
                a_3 * (T_1 - T_1_ave) / sigma_T_1 * (pi - pi_ave) / sigma_pi +
                a_4 * (T_1 - T_1_ave) / sigma_T_1 +
                a_5 * (n - n_ave) / sigma_n +
                a_6 * ((n - n_ave) / sigma_n) ** 2
        )

    def get_eta_isentropic(self, p_outlet: float, inputs: Inputs) -> float:
        """
        Returns the isentropic efficiency based on the regressions of Mirko Engelpracht.

        Args:
            p_outlet (float): High pressure value.
            inputs (Inputs): Input parameters for the calculation.

        Returns:
            float: Isentropic efficiency.
        """
        # If not constant value is given, eta_is is calculated based on the regression of Mirko Engelpracht
        n = self.get_n_absolute(inputs.n)

        a_1 = 0.5816
        a_2 = 0.002604
        a_3 = -1.515e-7
        a_4 = -0.00473
        pi = p_outlet / self.state_inlet.p
        eta = (
                a_1 +
                a_2 * n +
                a_3 * n ** 3 +
                a_4 * pi ** 2
        )
        if eta <= 0:
            raise ValueError("Efficiency is lower or equal to 0")
        return eta

    def get_eta_mech(self, inputs: Inputs) -> float:
        """
        Returns the mechanical efficiency based on the regressions of Mirko Engelpracht.

        Args:
            inputs (Inputs): Input parameters for the calculation.

        Returns:
            float: Mechanical efficiency.
        """
        p_outlet = self.get_p_outlet()
        n = self.get_n_absolute(inputs.n)
        # If not constant value is given, eta_is is calculated based on the regression of Mirko Engelpracht
        a_00 = 0.2199
        a_10 = -0.0193
        a_01 = 0.02503
        a_11 = 8.817e-5
        a_20 = -0.001345
        a_02 = -0.0003382
        a_21 = 1.584e-5
        a_12 = -1.083e-6
        a_22 = -5.321e-8
        a_03 = 1.976e-6
        a_13 = 4.053e-9
        a_04 = -4.292e-9
        pi = p_outlet / self.state_inlet.p
        return (
                a_00 +
                a_10 * pi + a_01 * n + a_11 * pi * n +
                a_20 * pi ** 2 + a_02 * n ** 2 + a_21 * pi ** 2 * n + a_12 * pi * n ** 2 + a_22 * pi ** 2 * n ** 2 +
                a_03 * n ** 3 + a_13 * pi * n ** 3 +
                a_04 * n ** 4
        )



import logging
import abc

import numpy as np
from vclibpy.components.heat_exchangers.heat_exchanger import HeatExchanger


logger = logging.getLogger(__name__)


class BasicNTU(HeatExchanger, abc.ABC):
    """
    Moving boundary NTU based heat exchanger.

    See parent class for more arguments.

    Args:
        flow_type (str):
            Counter, Cross or parallel flow
        ratio_outer_to_inner_area (float):
            The NTU method uses the overall heat transfer coefficient `k`
            and multiplies it with the outer area `A` (area of the secondary side).
            However, depending on the heat exchanger type, the areas may
            differ drastically. For instance in an air-to-water heat exchanger.
            The VDI-Atlas proposes the ratio of outer area to inner pipe area
            to account for this mismatch in sizes.
            The calculation follows the code in the function `calc_k`.
            Typical values are around 20-30.
    """

    def __init__(self, flow_type: str, ratio_outer_to_inner_area: float, **kwargs):
        """
        Initializes BasicNTU.

        Args:
            flow_type (str): Type of flow: Counter, Cross, or Parallel.
            ratio_outer_to_inner_area (float):
                The NTU method uses the overall heat transfer coefficient `k`
                and multiplies it with the overall area `A`.
                However, depending on the heat exchanger type, the areas may
                differ drastically. For instance in an air-to-water heat exchanger.
                The VDI-Atlas proposes the ratio of outer area to inner pipe area
                to account for this mismatch in sizes.
                The calculation follows the code in the function `calc_k`.
            **kwargs: Additional keyword arguments passed to the parent class.
        """
        super().__init__(**kwargs)
        self.ratio_outer_to_inner_area = ratio_outer_to_inner_area

        # Set primary cp:
        self._primary_cp = None

        # Type of HE:
        self.flow_type = flow_type.lower()
        if self.flow_type not in ["counter", "cross", "parallel"]:
            raise TypeError("Given flow_type is not supported")

    def set_primary_cp(self, cp: float):
        """
        Set the specific heat (cp) for the primary medium.

        Args:
            cp (float): Specific heat for the primary medium.
        """
        self._primary_cp = cp

    def calc_eps(self, R: float, NTU: float) -> float:
        """
        Calculate the effectiveness (eps) of the heat exchanger based on NTU.

        Source of implementation: EBC Lecture SimModelle.

        Args:
            R (float): Ratio of heat capacity rates (m_flow*cp) of the primary to secondary medium.
            NTU (float): Number of Transfer Units.

        Returns:
            float: Effectiveness (eps) of the heat exchanger.
        """
        if R in (0, 1):
            return NTU / (NTU + 1)
        if self.flow_type == "counter":
            return (1 - np.exp(-NTU * (1 - R))) / (1 - R * np.exp(-NTU * (1 - R)))
        if self.flow_type == "cross":
            if NTU == 0:
                return 0
            eta = NTU ** -0.22
            return 1 - np.exp((np.exp(- NTU * R * eta) - 1) / (R * eta))
        if self.flow_type == "parallel":
            return (1 - np.exp(-NTU * (1 + R))) / (1 + R)
        raise TypeError(f"Flow type {self.flow_type} not supported")

    def calc_R(self) -> float:
        """
        Calculate the R value, which is the ratio of heat capacity rates (m_flow*cp) of the primary to secondary medium.

        Returns:
            float: R value.
        """
        m_flow_pri_cp = self.m_flow * self._primary_cp
        if m_flow_pri_cp > self.m_flow_secondary_cp:
            return self.m_flow_secondary_cp / m_flow_pri_cp
        return m_flow_pri_cp / self.m_flow_secondary_cp

    def calc_k(self, alpha_pri: float, alpha_sec: float) -> float:
        """
        Calculate the overall heat transfer coefficient (k) of the heat exchanger.

        Args:
            alpha_pri (float): Heat transfer coefficient for the primary medium.
            alpha_sec (float): Heat transfer coefficient for the secondary medium.

        Returns:
            float: Overall heat transfer coefficient (k).
        """
        k_wall = self.calc_wall_heat_transfer()
        k = (1 / (
                        (1 / alpha_pri) * self.ratio_outer_to_inner_area +
                        (1 / k_wall) * self.ratio_outer_to_inner_area +
                        (1 / alpha_sec)
                )
             )
        return k

    @staticmethod
    def calc_NTU(A: float, k: float, m_flow_cp: float) -> float:
        """
        Calculate the Number of Transfer Units (NTU) for the heat exchanger.

        Args:
            A (float): Area of the heat exchanger.
            k (float): Overall heat transfer coefficient.
            m_flow_cp (float): Minimal heat capacity rates (m_flow*cp) between primary and secondary side.

        Returns:
            float: Number of Transfer Units (NTU).
        """
        return k * A / m_flow_cp

    def calc_m_flow_cp_min(self) -> float:
        """
        Calculate the minimum value between the heat capacity rates (m_flow*cp) for the primary and secondary mediums.

        Returns:
            float: Minimum value.
        """
        return min(
            self.m_flow * self._primary_cp,
            self.m_flow_secondary_cp
        )

    def calc_Q_ntu(self, dT_max: float, alpha_pri: float, alpha_sec: float, A: float) -> (float, float):
        """
        Calculate the heat transfer and overall heat transfer coefficient for the heat exchanger based on NTU.

        Args:
            dT_max (float): Maximum temperature differential.
            alpha_pri (float): Heat transfer coefficient for the primary medium.
            alpha_sec (float): Heat transfer coefficient for the secondary medium.
            A (float): Area of the heat exchanger.

        Returns:
            Tuple[float, float]: Heat transfer and overall heat transfer coefficient.
        """
        R = self.calc_R()
        k = self.calc_k(alpha_pri, alpha_sec)
        m_flow_cp_min = self.calc_m_flow_cp_min()
        NTU = self.calc_NTU(A, k, m_flow_cp_min)
        eps = self.calc_eps(R, NTU)

        # Get the maximal allowed heat flow
        Q_max = m_flow_cp_min * dT_max
        return Q_max * eps, k



"""
Module with classes for EV models.
They are not used for the calculation.
They may be used to see if the output is correct.
"""
import abc

from vclibpy.components.component import BaseComponent


class ExpansionValve(BaseComponent, abc.ABC):
    """Base class for an expansion valve.

    Args:
        A (float): Cross-sectional area of the expansion valve.
    """

    def __init__(self, A):
        super().__init__()
        self.A = A  # Cross-sectional area of the expansion valve

    @abc.abstractmethod
    def calc_m_flow_at_opening(self, opening) -> float:
        """
        Calculate the mass flow rate for the given opening.

        Args:
            opening (float): Opening of valve between 0 and 1

        Returns:
            float: Mass flow rate in kg/s
        """
        raise NotImplementedError

    @abc.abstractmethod
    def calc_opening_at_m_flow(self, m_flow, **kwargs) -> float:
        """
        Calculate the opening for the given mass flow rate

        Args:
            m_flow (float): Mass flow rate in kg/s
            **kwargs: Possible keyword arguments for child classes

        Returns:
            float: Opening
        """
        raise NotImplementedError

    def calc_outlet(self, p_outlet: float):
        """
        Calculate isenthalpic expansion valve.

        Args:
            p_outlet (float): Outlet pressure level
        """
        self.state_outlet = self.med_prop.calc_state("PH", p_outlet, self.state_inlet.h)



import logging

from vclibpy.components.heat_exchangers.ntu import BasicNTU
from vclibpy.media import ThermodynamicState

logger = logging.getLogger(__name__)


class VaporInjectionEconomizerNTU(BasicNTU):
    """
    Economizer heat exchanger which is NTU based.
    Used only for vapor injection cycles, as
    calculations are purely based for two-phase
    and liquid estimations.

    See parent class for more arguments.

    Assumptions:

    - Default `flow_type` is counter_flow.
    - Default `ratio_outer_to_inner_area` is 1, as
    - pipes are nearly same diameter and length
    - Secondary heat transfer for alpha is disabled; gas,
    liquid and two-phase models are used for both sides.
    """

    def __init__(self, **kwargs):
        kwargs.pop("secondary_heat_transfer", None)
        kwargs.pop("secondary_medium", None)
        self._state_two_phase_outlet = None
        self._state_two_phase_inlet = None
        super().__init__(
            flow_type=kwargs.pop("flow_type", "counter"),
            secondary_heat_transfer="None",
            secondary_medium="None",
            ratio_outer_to_inner_area=kwargs.pop("ratio_outer_to_inner_area", 1),
            **kwargs)

    @property
    def state_two_phase_inlet(self) -> ThermodynamicState:
        return self._state_two_phase_inlet

    @state_two_phase_inlet.setter
    def state_two_phase_inlet(self, state_inlet: ThermodynamicState):
        self._state_two_phase_inlet = state_inlet

    @property
    def state_two_phase_outlet(self) -> ThermodynamicState:
        return self._state_two_phase_outlet

    @state_two_phase_outlet.setter
    def state_two_phase_outlet(self, state_outlet: ThermodynamicState):
        self._state_two_phase_outlet = state_outlet

    def calc(self, inputs, fs_state) -> (float, float):
        raise NotImplementedError("Could be moved from VaporInjectionEconomizer")

    def set_secondary_cp(self, cp: float):
        """Set primary m_flow_cp"""
        self._secondary_cp = cp

    def start_secondary_med_prop(self):
        self.med_prop_sec = self.med_prop

    def terminate_secondary_med_prop(self):
        pass  # Not required as it is the central med_prop class

    def calc_alpha_secondary(self, transport_properties):
        raise NotImplementedError("Economizer does not use secondary heat transfer model.")

    def calc_transport_properties_secondary_medium(self, T, p=None):
        raise NotImplementedError("Economizer does not use this method")



from .states import ThermodynamicState, TransportProperties
from .media import get_two_phase_limits, MedProp
from .cool_prop import CoolProp
from .ref_prop import RefProp


__all__ = ['ThermodynamicState',
           'TransportProperties',
           'MedProp',
           'CoolProp',
           'RefProp']

USED_MED_PROP = (CoolProp, {})


def set_global_media_properties(med_prop_class: object, **kwargs):
    """
    Set the globally used MedProp class.

    Args:
        med_prop_class (object):
            Available MedProp children class.
        kwargs (dict):
            Additional settings for the MedProp class,
            e.g. {"use_high_level_api": True} for CoolProp.
    """
    global USED_MED_PROP
    USED_MED_PROP = (med_prop_class, kwargs)


def get_global_med_prop_and_kwargs():
    """
    Get the global MedProp class used
    for all calculations.
    Returns:
        MedProp: The class
    """
    global USED_MED_PROP
    return USED_MED_PROP[0], USED_MED_PROP[1]



# -*- coding: utf-8 -*-
"""
Created on 24.04.2023

@author: Christoph Hoeges

Last Update: 24.04.2023
"""
import pandas as pd


def print_states(**kwargs):
    """
    Transforms given states to DataFrame and prints table layout.
    Declaration of state must start with 'state'

    Returns:
    :return pandas.DataFrame df_states:
        DataFrame with states
    """
    # Remove keys in given data that are not label as "state"
    for key in kwargs.keys():
        if not key.startswith("state"):
            print(
                f"Given value {key} is removed from data transformation! "
                f"Value must be declared as 'state'"
            )
            kwargs.pop(key, None)

    # Extract data from ThermodynamicStates
    data_states = {key: {"state": key[5:], "T": data.T, "p": data.p, "h": data.h, "s": data.s, "d": data.d}
                   for key, data in kwargs.items()}
    df_states = pd.DataFrame.from_dict(data=data_states)

    # Similar to previous results - now with rounded data
    data_states_rounded = {key: {"State": key[5:],
                                 "T in °C": round(data.T - 273.15, 2),
                                 "p in bar": round(data.p * 1e-5, 3),
                                 "h in kJ/kg": round(data.h * 1e-3, 3),
                                 "s in kJ/kg": round(data.s * 1e-3, 3),
                                 "d in kg/m3": round(data.d, 3)}
                           for key, data in kwargs.items()}
    df_states_rounded = pd.DataFrame.from_dict(data=data_states_rounded)
    print(df_states_rounded.T)

    return df_states



# -*- coding: utf-8 -*-
"""
Created on 21.04.2020

@author: Christoph Hoeges, Fabian Wuellhorst, Jona Brach

To test:
- Transport properties are not fully tested (Status: 11.06.2020)
- Error raising is not implemented at all refprop calls
    - Additional change might be that not all errors and warning are excluded when 'use_error_check' is set to false but
        only warnings or errors?
"""
import logging
import os
import warnings
import shutil
import atexit

from ctREFPROP.ctREFPROP import REFPROPFunctionLibrary

from vclibpy.media import ThermodynamicState, TransportProperties, MedProp


logger = logging.getLogger(__name__)


class RefProp(MedProp):
    """
    Class to connect to refProp package.

    Args:
    :param string fluid_name:
        Fluid name for RefProp use
    :param list or None z:
        Fluid composition. Should only be used, when a self-design mixture shall be used. Further information
        see notes.
        When you want to use a self-design mixture to as follows:
        - Fluid-name needs to contain all component names within mixture: "R32.FLD|R125.FLD"
        - z needs to be a list with molar fractions: [0.697, 0.303]
        - Used example would be similar to R410A

    :param string dll_path:
        Specifier for the dll path used for RefProp,
        e.g. dll_path='C:\\path_to_dll\\RefProp64.dll'.
        If None, the `ref_prop_path` and function `get_dll_path` are
        used to determine the dll path
    :param boolean use_error_check:
        Specifier whether errors and warnings shall be checked when calling RefProp or not
    :param boolean use_warnings:
        Specifier whether warnings shall be used
    :param str ref_prop_path:
        Path to RefProp package. Default is the ENV variable `RPPREFIX`.
    :param bool copy_dll:
        If True (not the default), a copy of the dll is created to enable simultaneous use of
        multiple fluids in multiprocessing.
    :param str copy_dll_directory:
        If `copy_dll` is True, the DLL is copied to this directory.
        If None (default), the current working directory is used.

    Note:
        - You need to install package ctREFPROP Package
          https://github.com/usnistgov/REFPROP-wrappers/tree/master/wrappers/python
        - In case you use a self-defined mixture: Entropy reference state might deviate from GUI!!

    Functionality:
        - It does not work to have multiple instances simultaneously. When calculating values the last fluid name
            somehow will be used even though instance includes "new" name. Need to be fixed at some point.


    How to use RefProp-Wrapper:
    ---------------------------
    1.) Create RefProp instance: rp = RefProp("R32")
    2.) In case you want to calculate fluid properties (state variables) for a specific state: Use calc_state() function
            Multiple inputs can be given (but you need to now two in order to define the values). For further
            information see function header.
    3.) Further get-Functions implemented
        - get_gwp(): Global warming potential
        - get_odp(): Ozone depletion potential
        - get_safety(): Safety class
        - get_mass_fraction(): Mass fractions of pure substances in fluid
        - get_molar_mass(): Molar mass of fluid
        - get_mol_fraction(): Mol fractions of pure substances in fluid
        - get_comp_names(): Pure substances names in fluid
        - get_longname(): Long name of fluid within RefProp
        - get_critical_point(): Crit. Temperature and Pressure
        - get_version(): Version of wrapper and of RefProp dll


    Version notes:
    --------------
    0.1.0 (21.04.2020, Christoph Hoeges):
        First implementation
            - Contains multiple functions to call RefProp instance
            - Can calculate state, crit, GWP, ODP, Safety class, molar mass etc. of fluid
            - Mixtures might work but it wasn't fully testet - R455A e.g. still deviates slightly

    0.1.1 (25.04.2020, Christoph Hoeges):
        Multiple changes, added functionality. Commands are still the same though
            - Self designed mixtures are possible now as well (see instructions in init
            - Added composition input to __init__ function
            - Added modes in calc_state function and removed bug in PH calculation
            - Additional protected functions due to different input
            - Adjusted _call_refprop function (z is not mass frac but mol fraction)
            - Added documentation / instruction

    0.1.2 (08.05.2020, Christoph Hoeges):
        Multiple adjustments
            - Added function to call ABFLSHdll function in refprop (high level function)
            - Changed function calls in calc_state function to ABFLSH and adjusted conversion
            - Change init-function so user can choose which dll shall be used
            - Add function to get version of this current wrapper as well as RefProp-dll version

    0.1.3 (12.05.2020, Christoph Hoeges):
        Multiple changes
            - Added function to get all files in MIXTURE and FLUIDS directory to check available fluids
            - Added error function in order to return errors when RefProp-Functions are called.
                NOTE: Not all instances where refprop is called are checked for errors. Currently, it is only used in
                    init and calc_state

    0.1.4 (19.05.2020, Christoph Hoeges):
        Multiple changes:
            - Debugged fluid properties calculation for predefined mixtures
            - Fixed option of self-defined mixtures

    0.1.5 (22.05.2020, Christoph Hoeges):
        Include transport properties calculation into wrapper

    0.1.6 (10.06.2020, Fabian Wuellhorst):
        Add option to use a custom dll path. This necessary to use multiple instances with possible
        different fluids at the same time.
    """
    # General information
    #
    __version__ = "0.1.6"
    __author__ = "Christoph Hoeges"

    _fluid_mapper = {'air': 'air.ppf'}

    def __init__(self,
                 fluid_name,
                 z=None,
                 dll_path: str = None,
                 use_error_check: bool = True,
                 use_warnings: bool = True,
                 ref_prop_path: str = None,
                 copy_dll: bool = True,
                 copy_dll_directory: str = None
                 ):
        if ref_prop_path is None:
            # Get environment variable for path to dll
            ref_prop_path = os.environ["RPPREFIX"]
        if dll_path is None:
            path_to_dll = self.get_dll_path(ref_prop_path)
        else:
            path_to_dll = dll_path

        if copy_dll:
            if copy_dll_directory is None:
                copy_dll_directory = os.getcwd()
            try:
                self._delete_dll_path = os.path.join(
                    copy_dll_directory,
                    f"med_prop_{fluid_name}_REFPRP64.dll"
                )
                shutil.copyfile(path_to_dll, self._delete_dll_path)
                atexit.register(self.terminate)
                path_to_dll = self._delete_dll_path
            except (PermissionError, FileExistsError) as err:
                logger.error("Can't copy file to new path: %s", err)
        else:
            self._delete_dll_path = None
        logger.info("Using dll: %s", path_to_dll)

        super().__init__(fluid_name=fluid_name)

        self._flag_check_errors = use_error_check
        self._flag_warnings = use_warnings
        # Set path to RefProp package
        self._ref_prop_path = ref_prop_path
        self.rp = REFPROPFunctionLibrary(path_to_dll)
        self.rp.SETPATHdll(ref_prop_path)
        self.molar_base_si = self.rp.GETENUMdll(0, "MOLAR BASE SI").iEnum
        # Set fluid name
        self.fluid_name = fluid_name
        # Get mass and mol fraction and number of components
        self._get_comp_frac(z)
        # Get component names
        self._comp_names = self._get_comp_names()
        # Mixture flag
        if self._n_comp > 1:
            self._mix_flag = True
        else:
            self._mix_flag = False

        # Setup
        self._setup_rp()
        # Calculate molar mass in kg/mol
        # self.M = self._call_refprop_allprop("M").Output[0]
        self.M = self._call_refprop(inp_name="", out_name="M").Output[0]  # kg/mol

        self._nbp = None

    def terminate(self):
        if self._delete_dll_path is not None:
            self._delete_dll()

    def _delete_dll(self):
        try:
            # Taken from here: https://stackoverflow.com/questions/21770419/free-the-opened-ctypes-library-in-python
            import _ctypes
            import sys
            _handle = self.rp.dll._handle
            if sys.platform.startswith('win'):
                _ctypes.FreeLibrary(_handle)
            else:
                _ctypes.dlclose(_handle)
            os.remove(self._delete_dll_path)
            self._delete_dll_path = None
        except (FileNotFoundError, PermissionError) as err:
            logger.error(
                "Could not automatically delete the copied RefProp dll at %s. "
                "Delete it yourself! Error message: %s", self._delete_dll_path, err
            )

    def _call_refprop_abflsh(self,
                             inp_name,
                             value_a,
                             value_b,
                             i_flag=1):
        """ Call RefProp via ABFLSHdll method
        You can define multiple inputs but only "specific ones" where no input values are needed for
        e.g. M, Tcrit, pcrit

        Parameters:
        -----------
        :param string inp_name:
            Input commands: "PQ"
        :param float value_a:
            Value of parameter b defined in inp_name. In case of None 0 will be used.
        :param float value_b:
            Value of parameter b defined in inp_name. In case of None 0 will be used.
        :param int i_flag:
            Flag
        Return:
        -------
        :return ABFLSHdlloutput tmp:
            Returns ABFLSH output

        """
        # TODO
        tmp = self.rp.ABFLSHdll(inp_name, value_a, value_b, self._mol_frac, i_flag)

        return tmp

    def _call_refprop_allprop(self,
                              out_name,
                              T_val=None,
                              d_val=None,
                              i_mass=0,
                              i_flag=1):
        """ Call RefProp via ALLPROPSdll-method

        Parameters:
        -----------
        :param string out_name:
            Variables you want to calculate. Multiple outputs are possible:
                - Single: "M"
                - Multiple: "M,TC,PC"
        :param float T_val:
            Temperature in current state in K.
            Note: In case you want to get general fluid parameters such as M, Tcrit, .. Stick to default value!
        :param float d_val:
            Density in current state (unit depending on i_mass flag - either mol/m^3 or kg/m^3)
            Note: In case you want to get general fluid parameters such as M, Tcrit, .. Stick to default value!
        :param int i_mass:
            Specifies which units the inputs are given in.
                - 0: Molar based
                - 1: Mass based
            Note: In current version (10.0.0.72) Ian Bell says in multiple Git-Issues that you should stick to molar
                    base!
        :param int i_flag:
            In current version (10.0.0.72) i_flag is used to define whether a string containing the units is written in
            'hUnits'.
                - 0: Deactivated (increases the calculation speed)
                - 1: activated
        Return:
        -------
        :return ALLPROPSdlloutput result:
            List with values for parameters

        """
        # Check values of T and d
        if T_val is None:
            T_val = 0
        if d_val is None:
            d_val = 0
        # Define fraction used depending on i_mass flag
        if i_mass == 0:
            frac = self._mol_frac
        elif i_mass == 1:
            frac = self._mass_frac
        else:
            raise ValueError("Chosen i_mass flag '{}' is not possible in ALLPROPSdll function!".format(i_mass))

        # Call RefProp
        res = self.rp.ALLPROPSdll(out_name, self.molar_base_si, i_mass, i_flag, T_val, d_val, frac)

        return res

    def _call_refprop(self,
                      inp_name,
                      out_name,
                      value_a=None,
                      value_b=None,
                      i_mass=0,
                      i_flag=1,
                      frac=None,
                      fluid=None):
        """ Call general refProp function and calculate values

        Parameters:
        -----------
        :param string fluid:
            Fluid name - in case default value None is used, stored fluid name will be used for command
        :param string inp_name:
            Input parameter specification
        :param string out_name:
            Output string name
        :param float value_a:
            Value of parameter b defined in inp_name. In case of None 0 will be used.
        :param float value_b:
            Value of parameter b defined in inp_name. In case of None 0 will be used.
        :param integer i_flag:
            Defines further settings (see documentation)
        :param int i_mass:
            Specifies which units the inputs are given in.  # TODO: WRONG! iMass determines composition, iUnits determines properties (except q)
                - 0: Molar based
                - 1: Mass based
            Note: In current version (10.0.0.72) Ian Bell says in multiple Git-Issues that you should stick to molar
                    base!
        :param list frac:
            List with either mol or mass fraction of pure substances in current fluid
                (in case of single pure substance: [1]).

        Return:
        -------
        :return REFPROPdlloutput output:
            Command of refprop
        """
        # Check inputs
        if value_a is None:
            value_a = 0
        if value_b is None:
            value_b = 0

        if fluid is None:
            if self._predefined:
                fluid = self.fluid_name  # TODO: in general not necessary, decreases performance
            else:
                fluid = ""
        if frac is None:
            if self._predefined:
                frac = ""
            else:
                if i_mass == 0:
                    frac = self._mol_frac
                elif i_mass == 1:
                    frac = self._mass_frac
                else:
                    raise ValueError("Variable i_mass has invalid input '{}'".format(i_mass))

        # Call refprop function
        tmp = self.rp.REFPROPdll(fluid,
                                 inp_name,
                                 out_name,
                                 self.molar_base_si,
                                 i_mass,
                                 i_flag,
                                 value_a,
                                 value_b,
                                 frac)

        return tmp

    def _check_error(self,
                     err_num,
                     err_msg,
                     func_name=""):
        """ Check error code and raise error in case it is critical

        Parameters:
        -----------
        :param integer err_num:
            Error return code
        :param string err_msg:
            Error message given in RefProp call
        :param string func_name:
            Name of function error needs to be checked in
        """
        # All fine in case error number is 0
        # Throw warning in case error number different than 0 and smaller than 100 is given
        # Throw error in case error number higher than 100 is given
        if err_num:
            if err_num < 100:
                if self._flag_warnings:
                    warnings.warn("[WARNING] Error number {} was given in function '{}'. No critical error but "
                                  "something went wrong maybe. \n Error message is: '{}'".format(str(err_num),
                                                                                                 func_name, err_msg))
            else:
                if self._flag_check_errors:
                    raise TypeError("[ERROR] When calling RefProp in function '{}' error number {} was "
                                    "returned. \n Error message is: '{}'".format(func_name, str(err_num), err_msg))

    def _get_comp_names(self):
        """ Get component names. In case current fluid is mixture, only component names will be returned.
        In case fluid is a pure substance, substance name is returned.

        Return:
        -------
        :return list comp_names:
            List with pure substances in current refrigerant
        """
        comp_names = []
        if self._predefined:
            # Note: While trying it was possible to get fluid name as well, therefore n_comp+1 is used.
            if self._n_comp > 1:
                for i in range(1, self._n_comp + 3):
                    test = self.rp.NAMEdll(i)
                    tmp = test.hn80.replace(".FLD", "")
                    if not tmp == "":
                        comp_names.append(tmp)
            else:
                for i in range(self._n_comp + 3):
                    tmp = self.rp.NAMEdll(i).hnam
                    if not tmp == "":
                        comp_names.append(tmp)

        else:
            # Self-defined
            tmp_str = self.fluid_name.split("|")
            for i in tmp_str:
                i = i.replace(".FLD", "")
                i = i.replace(".MIX", "")
                if len(i) < 2:
                    continue
                else:
                    comp_names.append(i)

        # Replace Carbon Dioxide for CO2
        for i, tmp in enumerate(comp_names):
            if "Carbon dio" in tmp:
                comp_names[i] = "CO2"

        return comp_names

    def _get_comp_frac(self, z):
        """ Get mass/mol fraction and number of components of current fluid

        Parameters:
        -----------
        :param list z:
            Contains predefined molar fractions in case one is given. Otherwise, z will be None
        """
        # Check if predefined or not
        # Predefined
        if z is None:
            self._predefined = True
        # Pure substance
        elif len(z) == 1:
            self._predefined = True
        # Self-designed mixture
        else:
            self._predefined = False

        # In case predefined mixture or pure substance is used
        if self._predefined:
            # Dummy function to get values for z in order to specify number of components
            tmp_mol = self.rp.REFPROPdll(self.fluid_name, "PQ", "H", self.molar_base_si, 0, 0, 101325, 0, [1])
            tmp_mass = self.rp.REFPROPdll(self.fluid_name, "PQ", "H", self.molar_base_si, 1, 0, 101325, 0, [])
            # Check for errors
            self._check_error(tmp_mol.ierr, tmp_mol.herr, self._get_comp_frac.__name__)
            self._check_error(tmp_mass.ierr, tmp_mass.herr, self._get_comp_frac.__name__)
            # Mass and molar fractions of components
            self._mol_frac = [zi for zi in tmp_mol.z if zi > 0]
            self._mass_frac = [zi for zi in tmp_mass.z if zi > 0]
            # Get number of components
            self._n_comp = len(self._mol_frac)
            # Check, whether error occurred
            if self._n_comp < 1:
                # It might be possible that z value bugs when calling RefProp. In case of a pure substance this does not
                # matter so an additional filter is included
                if len(self._mass_frac) == 1:
                    self._n_comp = 1
                    self._mol_frac = [1]
                else:
                    raise ValueError("Number of components for current fluid '{}' is less than "
                                     "one!".format(self.fluid_name))

            # Get mol fraction
            # self._mol_frac = self._transform_to_molfraction(self._mass_frac)
        else:
            # Mol fraction
            self._mol_frac = z
            self._mass_frac = []
            # Get number of components
            self._n_comp = len(self._mol_frac)

    def _setup_rp(self):
        """ Setup for RefProp """
        # Errors can occur in case REFPROP is initalizated multiple time with same fluid - thus a pre setup is used here
        self.rp.SETUPdll(1, "N2", "HMX.BNC", "DEF")
        # In case of pure substance
        if self._n_comp == 1:
            self.rp.SETUPdll(self._n_comp, self.fluid_name, "HMX.BNC", "DEF")
        # In case of mixtures
        else:
            # Check if mixture is predefined
            if self._predefined:
                # Pre defined mixture - different baseline operating point is used
                mode = 2
                mixture = "|".join([f+".FLD" for f in self._comp_names])
                n_comp = self._n_comp
            else:
                # Self defined mixture
                mode = 1
                n_comp = self._n_comp
                # Define mixtures name
                # TODO: Ending is not necessary to create mixtures....
                mixture = "|".join([f+".FLD" for f in self._comp_names])

            # Setup for mixture
            setup = self.rp.SETUPdll(n_comp, mixture, 'HMX.BNC', 'DEF')
            setref = self.rp.SETREFdll("DEF", mode, self._mol_frac, 0, 0, 0, 0)
            # z = self._mol_frac
            # Get mass fraction
            self._mass_frac = self._transform_to_massfraction(self._mol_frac)

            # Check whether mixing rules are available
            if setup.ierr == 117:
                if self._flag_check_errors:
                    raise ValueError(
                        "[MIXING ERROR] Mixing rules for mixture '{}' do not exist!".format(self._comp_names))
                else:
                    print(
                        "[MIXING ERROR] Mixing rules for mixture '{}' do not exist!".format(self._comp_names))
            elif setup.ierr == -117:
                if self._flag_warnings:
                    warnings.warn(
                        "[MIXING ERROR] Mixing rules for mixture '{}' are estimated!".format(self._comp_names))
                else:
                    print(
                        "[MIXING WARNING] Mixing rules for mixture '{}' are estimated!".format(self._comp_names))

    def _transform_to_massfraction(self,
                                   mol_frac):
        """ Transforms mol fraction to mass fraction

        Parameters:
        -----------
        :param list mol_frac:
            List containing floats for mol fraction

        Return:
        -------
        :return list mass_frac:
            List containing floats for mass fraction
        """
        tmp = self.rp.XMASSdll(mol_frac)
        mass_frac = [yi for yi in tmp.xkg if yi > 0]
        return mass_frac

    def _transform_to_molfraction(self,
                                  mass_frac):
        """ Transforms mass fraction to mol fraction

        Parameters:
        -----------
        :param list mass_frac:
            List containing floats for mass fraction

        Return:
        -------
        :return list frac:
            List containing floats for mol fraction
        """
        tmp = self.rp.XMOLEdll(mass_frac)
        mol_frac = [xi for xi in tmp.xmol if xi > 0]
        return mol_frac

    def calc_state(self, mode: str, var1: float, var2: float, kr=1):
        """ Calculate state. Depending on mode, different function will be chosen. Input state variables need to be in
        SI units!

        Notes:
        ------
        1.) PT does not work when state might be within the two-phase region!
        2.) Only functions for density are implemented. In case you know the specific volume instead use density
                functions with inverse value!
        3.) Quality can have values outside of 'physical' scope:
                q = -998: Subcooled liquid
                q = 998: Superheated vapor
                q = 999: Supercritical state

        Possible modes are currently:
            - "PD": Pressure, Density
            - "PH": Pressure, Enthalpy
            - "PQ": Pressure, Quality
            - "PS": Pressure, Entropy
            - "PT": Pressure, Temperature
            - "PU": Pressure, Internal Energy
            - "TD": Temperature, Density
            - "TH": Temperature, Enthalpy
            - "TQ": Temperature, Quality
            - "TS": Temperature, Entropy
            - "TU": Temperature, Internal Energy
            - "DH": Density, Enthalpy
            - "DS": Density, Entropy
            - "DU": Density, Internal Energy
            - "HS": Enthalpy, Entropy

        Parameters:
        -----------
        :param string mode:
            Defines which input state variables are given (see possible modes above)
        :param float var1:
            Value of state variable 1 (first one in name) - use SI units!
        :param float var2:
            Value of state variable 2 (second one in name) - use SI units!
        :param int kr:
            phase flag (kr=1: lower density, kr=2: higher density)
            relevant for "TH", "TS", "TU"

        Return:
        -------
        :return ThermodynamicState state:
            Thermodynamic state with state variables
        """
        # Multiplier for pressure since kPa is used in RefProp
        p_multi = 1e-3
        # Multiplier for energy
        e_multi = self.M
        # Multiplier for density
        d_multi = 1 / self.M / 1000

        # Init all parameters
        p = None
        d = None
        T = None
        u = None
        h = None
        s = None
        q = None

        # Check modi

        # Pressure and density
        if mode == "PD":
            p = var1
            d = var2
            var1 = var1 * p_multi
            var2 = var2 * d_multi
            tmp = self.rp.PDFLSHdll(var1, var2, self._mol_frac)
        # Pressure and enthalpy
        elif mode == "PH":
            p = var1
            var1 = var1 * p_multi
            h = var2
            var2 = var2 * e_multi
            tmp = self.rp.PHFLSHdll(var1, var2, self._mol_frac)
        # Pressure and quality
        elif mode == "PQ":
            p = var1
            var1 = var1 * p_multi
            q = var2
            # In case current fluid is mixture you need to transform Q to molar base for RefProp-function
            if self._mix_flag:
                var2 = self._call_refprop("PQMASS", "QMOLE", p, q, i_mass=1).Output[0]
            tmp = self.rp.PQFLSHdll(var1, var2, self._mol_frac, 0)
        # Pressure and entropy
        elif mode == "PS":
            p = var1
            var1 = var1 * p_multi
            s = var2
            var2 = var2 * e_multi
            tmp = self.rp.PSFLSHdll(var1, var2, self._mol_frac)
        # Pressure and Temperature
        elif mode == "PT":
            p = var1
            var1 = var1 * p_multi
            T = var2
            tmp = self.rp.TPFLSHdll(var2, var1, self._mol_frac)
        # Pressure and internal energy
        elif mode == "PU":
            p = var1
            var1 = var1 * p_multi
            u = var2
            var2 = var2 * e_multi
            # mode = "PE"
            tmp = self.rp.PEFLSHdll(var1, var2, self._mol_frac)
        # Temperature and density
        elif mode == "TD":
            T = var1
            d = var2
            var2 = var2 * d_multi
            tmp = self.rp.TDFLSHdll(var1, var2, self._mol_frac)
        # Temperature and enthalpy
        elif mode == "TH":
            T = var1
            h = var2
            var2 = var2 * e_multi
            tmp = self.rp.THFLSHdll(T, var2, self._mol_frac, kr)
        # Temperature and quality
        elif mode == "TQ":
            T = var1
            q = var2
            # In case current fluid is mixture you need to transform Q to molar base for RefProp-function
            if self._mix_flag:
                var2 = self._call_refprop("TQMASS", "QMOLE", T, q, i_mass=1).Output[0]
            tmp = self.rp.TQFLSHdll(T, var2, self._mol_frac, 1)
        # Temperature and entropy
        elif mode == "TS":
            T = var1
            s = var2
            var2 = var2 * e_multi
            tmp = self.rp.TSFLSHdll(T, var2, self._mol_frac, kr)
        # Temperature and internal energy
        elif mode == "TU":
            T = var1
            u = var2
            var2 = var2 * e_multi
            # mode = "TE"
            tmp = self.rp.TEFLSHdll(T, var2, self._mol_frac, kr)
        # Density and enthalpy
        elif mode == "DH":
            d = var1
            var1 = var1 * d_multi
            h = var2
            var2 = var2 * e_multi
            tmp = self.rp.DHFLSHdll(var1, var2, self._mol_frac)
        # Density and entropy
        elif mode == "DS":
            d = var1
            var1 = var1 * d_multi
            s = var2
            var2 = var2 * e_multi
            tmp = self.rp.DSFLSHdll(var1, var2, self._mol_frac)
        # Density and inner energy
        elif mode == "DU":
            d = var1
            var1 = var1 * d_multi
            u = var2
            var2 = var2 * e_multi
            # mode = "DE"
            tmp = self.rp.DEFLSHdll(var1, var2, self._mol_frac)
        elif mode == "HS":
            h = var1
            var1 = var1 * e_multi
            s = var2
            var2 = var2 * e_multi
            tmp = self.rp.HSFLSHdll(var1, var2, self._mol_frac)
        else:
            raise ValueError("Chosen mode is not available in refprop calc_state function!")

        # Check for errors
        self._check_error(tmp.ierr, tmp.herr, self.calc_state.__name__)

        # Get all state variables
        if p is None:
            p = tmp.P / p_multi
        if T is None:
            T = tmp.T
        if u is None:
            u = tmp.e / e_multi
        if h is None:
            h = tmp.h / e_multi
        if s is None:
            s = tmp.s / e_multi
        if d is None:
            d = tmp.D / d_multi
        if q is None:
            if self._mix_flag:
                # Transform current q (molar) to mass based quality
                tmp2 = self._call_refprop("PH", "QMASS", p, h * e_multi, i_mass=1)
                if tmp2.Output[0] < 0:
                    q_mass = tmp2.ierr
                else:
                    q_mass = tmp2.Output[0]
                q = q_mass
            else:
                q = tmp.q

        # # In case not in two phase region reset q to -1
        # if q > 1 or q < 0:
        #     q = -1

        # Define state
        state = ThermodynamicState(p=p, T=T, u=u, h=h, s=s, d=d, q=q)
        return state

    def calc_satliq_state(self, s):
        """s in kJ/kgK"""
        s = s * self.M * 1000  # kJ/kgK -> J/molK
        tmp = self.rp.SATSdll(s=s, z="", kph=1)
        self._check_error(tmp.ierr, tmp.herr, self.calc_satliq_state.__name__)
        if tmp.k1 != 1:
            raise TypeError
        p = tmp.P1 * 1000  # kPa -> Pa
        d = tmp.D1 * self.M * 1000  # mol/l -> kg/mol
        return self.calc_state("PD", p, d)

    def calc_transport_properties(self, state: ThermodynamicState):
        """ Calculate transport properties of RefProp fluid at given state

        Parameters:
        -----------
        :param ThermodynamicState state:
            Current thermodynamic state
        Return:
        -------
        :return TransportProperties props:
            Instance of TransportProperties
        """
        # Get properties
        tmp = self._call_refprop_allprop("PRANDTL,VIS,TCX,KV,CV,CP,BETA,STN,ACF", state.T, state.d / self.M, i_mass=0)
        # Create transport properties instance
        props = TransportProperties(lam=tmp.Output[2],
                                    dyn_vis=tmp.Output[1],
                                    kin_vis=tmp.Output[3],
                                    pr=tmp.Output[0],
                                    cp=tmp.Output[5] / self.M,
                                    cv=tmp.Output[4] / self.M,
                                    beta=tmp.Output[6],
                                    sur_ten=tmp.Output[7],
                                    ace_fac=tmp.Output[8],
                                    state=state)
        # Return props
        return props

    def get_available_substances(self,
                                 mode="all",
                                 include_ending=False,
                                 save_txt=False):
        """ Get all available RefProp fluids (mixtures and / or pure substances depending on mode)

        Parameters:
        -----------
        :param string mode:
            Mode defining which kind of fluids you want to have: 'pure': pure substances, 'mix': mixtures, 'all': all
        :param boolean include_ending:
            Defines, whether file ending shall be returned as well or not
        :param boolean save_txt:
            Defines, whether a text file with names shall be created in current working directory
        Return:
        -------
        :return list names:
            String list containing names of available fluids (depending on defined mode)
        """
        # Possible endings
        _endings = ["MIX", "FLD", "PPF"]

        # Folders where fluid data is located
        folders = [r"FLUIDS", r"MIXTURES"]

        # Define paths by mode
        if mode == "pure":
            paths = [os.path.join(self._ref_prop_path, folders[0])]
        elif mode == "mix":
            paths = [os.path.join(self._ref_prop_path, folders[1])]
        elif mode == "all":
            paths = [os.path.join(self._ref_prop_path, folders[0]),
                     os.path.join(self._ref_prop_path, folders[1])]
        else:
            raise ValueError("Chosen mode '{}' is not possible!".format(mode))

        # Get files in folders, remove ending and append to names
        names = []
        for p in paths:
            # Get all files in directory
            files = [f for f in os.listdir(p) if os.path.isfile(os.path.join(p, f))]
            # Remove ending
            if include_ending:
                tmp = [path for path in files if path.split(".")[1] in _endings]
            else:
                tmp = [path.split(".")[0] for path in files if path.split(".")[1] in _endings]
            # Append names to names list
            names.extend(tmp)

        # In case names shall be stored
        if save_txt:
            with open("available_fluids.txt", "w") as output:
                for i in names:
                    output.write("{} \n".format(i))

        if not names:
            raise ValueError("No fluids are in current RefProp directory. Check path '{}'!".format(self._ref_prop_path))

        # Return names
        return names

    def get_comp_names(self):
        """ Get name of components within current fluid"

        Return:
        -------
        :return list comp_names:
            String names of components within current fluid
        """
        return self._comp_names

    def get_nbp(self):
        """ Get normal boiling point (T @ q=0 and 1 bar)

        Return:
        -------
        :return float nbp:
            Normal boiling point of refrigerant in °C
        """
        if not self._nbp:
            self._nbp = self.calc_state("PQ", 1e5, 0.0).T - 273.15

        return self._nbp

    def get_molar_composition(self, state: ThermodynamicState, z_molar=None):
        """ Get composition on molar base. Liquid phase, vapor phase and total.

        :param ThermodynamicState state: the state whose compositions will be returned
        :param list z_molar: molar composition of fluid. In case of None, default value _mol_frac is used
        :return:
            - list x:
                composition of liquid phase
            - list y:
                composition of vapor phase
            - list z:
                composition in total
        """
        if z_molar is None:
            z = self._mol_frac
            M = self.M
        else:
            z = z_molar
            M = self.rp.REFPROPdll(hFld="",
                                   hIn="",
                                   hOut="M",
                                   iUnits=self.molar_base_si,
                                   iMass=0,
                                   iFlag=1,
                                   a=0,
                                   b=0,
                                   z=z_molar).Output[0]
        num_components = len(z)

        tmp = self.rp.TDFLSHdll(T=state.T,
                                D=state.d / M / 1000,
                                z=z)
        # TDFLSHdll is deprecated, use the following in future:
        # tmp = self.rp.ABFLSHdll(ab="TD",
        #                         a=state.T,
        #                         b=state.d / self.M / 1000,
        #                         z=self._mol_frac,
        #                         iFlag=0)  # molar units

        x = list(tmp.x[:num_components])
        y = list(tmp.y[:num_components])

        return x, y, z

    def get_critical_point(self):
        """ Get T and p of critical point

        Return:
        -------
        :return float Tc:
            Temperature at critical point in K
        :return float pc:
            Pressure at critical point in Pa
        :return float dc:
            Density at critical point in kg/m^3
        """
        mode = 2
        if mode == 1:
            tmp = self._call_refprop("CRIT", "T,P,D", i_mass=1)
            Tc = tmp.Output[0]
            pc = tmp.Output[1]
            dc = tmp.Output[2] * self.M
        else:
            res = self._call_refprop_allprop("TC,PC,DC")
            Tc = res.Output[0]
            pc = res.Output[1]
            dc = res.Output[2] * self.M
        return Tc, pc, dc

    #
    #
    def get_def_limits(self):
        """ Get limits of current ref prop fluid
        Limits contain Tmin, Tmax, Dmax and Pmax (Temperatures, density, pressure)

        :return dict limits:
            Dictionary with definition limits in RefProp. Contains min/max temperature, max density, max pressure.
        """
        tmp = self._call_refprop_allprop("TMIN,TMAX,DMAX,PMAX")
        limits = {"Tmin": tmp.Output[0],
                  "Tmax": tmp.Output[1],
                  "Dmax": tmp.Output[2] * self.M,
                  "Pmax": tmp.Output[3]}
        return limits

    @staticmethod
    def get_dll_path(ref_prop_path: str):
        """
        Return the location of the dll

        Return:
        :return: string path_to_dll:
            Path of a valid dll
        """
        path_to_dll = os.path.join(ref_prop_path, r"REFPRP64.DLL")

        # Check if dll actually exists:
        if not os.path.exists(path_to_dll):
            raise FileNotFoundError("Selected dll not found automatically. "
                                    "Please alter the local attribute or search for yourself.")

        return path_to_dll

    def get_fluid_name(self):
        """ Get fluid name.

        Return:
        -------
        :return string fluid_name:
            Fluid name
        """
        return self.fluid_name

    def get_gwp(self):
        """ Get gwp of current fluid from refProp

        Return:
        -------
        :return float gwp:
            Global warming potential of fluid. In case calculation failed, None will be returned.
        """
        # Call refProp
        tmp = self._call_refprop("",
                                 "GWP",
                                 i_mass=1)
        # Calculate gwp
        gwp = round(sum(max(tmp.Output[i], 0) * self._mass_frac[i] for i in range(self._n_comp)), 2)
        # In case GWP cannot be calculated
        if gwp < 0:
            gwp = 0
        return gwp

    def get_longname(self):
        """ Get longname of fluid

        Return:
        -------
        :return string longname:
            Name of current fluid in refprop - provides mass fractions and components as well (in case of mixture)
        """
        longname = self._call_refprop("", "LONGNAME(0)").hUnits
        return longname

    def get_mass_fraction(self, use_round=True):
        """ Get mass fractions of pure substances in current fluid

        Parameters:
        :param boolean use_round:
            Flag to define, whether the exact values or rounded values (by the fourth number) shall be used
        Return:
        -------
        :return list mass_frac:
            List of component mass fractions
        """
        if use_round:
            mass_frac = [round(i, 4) for i in self._mass_frac]
        else:
            mass_frac = self._mass_frac
        return mass_frac

    def get_molar_mass(self):
        """ Get molar mass of current fluid

        Return:
        -------
        :return float M:
            Molar mass of current fluid in kg/mol
        """
        return self.M

    def get_mol_fraction(self, use_round=True):
        """ Get mol fractions of pure substances in current fluid

        Parameters:
        :param boolean use_round:
            Flag to define, whether the exact values or rounded values (by the fourth number) shall be used
        Return:
        -------
        :return list frac:
            List of component mol fractions
        """
        if use_round:
            mol_frac = [round(i, 4) for i in self._mol_frac]
        else:
            mol_frac = self._mol_frac
        return mol_frac

    def get_odp(self):
        """ Calculate ozone depletion potential
        In case of mixtures: Maximum value of pure substances will be used

        Return:
        -------
        :return float odp:
            ODP of fluid. In case calculation failed, None will be returned.
        """
        # Call refProp
        tmp = self._call_refprop("",
                                 "ODP",
                                 i_mass=1)
        # Calculate odp
        odp = max(max(tmp.Output), 0)
        # In case some error occured
        if odp < 0:
            odp = 0
        return odp

    def get_safety(self):
        """ Calculate safety class of refrigerant

        Return:
        -------
        :return string safety:
            Safety class of fluid.
        """
        # Call refProp
        tmp = self._call_refprop("",
                                 "SAFETY",
                                 i_mass=1)
        # Get safety class
        safety = tmp.hUnits
        # Return safety
        return safety

    def get_sat_vap_pressure(self, T_sat):
        """ Get pressure of saturated vapor for defined temperature

        Note:
            - works for vapor, liquid and solid
            - returns equilibrium pressure at defined line (q=1)

        Parameters:
        :param float T_sat:
            Temperature in K
        Return:
        :return float p_sat:
            Vapor pressure in Pa
        """
        trip = self.get_triple_point()
        if trip[0] <= T_sat:
            p_sat = self.calc_state("TQ", T_sat, 1.0).p
        else:
            tmp = self.rp.REFPROPdll("", "TSUBL", "P", self.molar_base_si, 0, 0, T_sat, 0.0, self._mol_frac)
            p_sat = tmp.Output[0]
        return p_sat

    def get_triple_point(self):
        """ Get temperature and pressure at triple point of current fluid

        Note: Works fine for pure substances, mixtures might not work properly

        Return:
        :return float T_tpl:
            Temperature at triple point in K
        :return float p_tpl:
            Pressure at triple point in Pa
        """
        # Call Refprop
        tmp = self._call_refprop("TRIP", "T;P")
        return tmp.Output[0], tmp.Output[1]

    def get_version(self):
        """ Get version of wrapper and used RefProp dll

        Return:
        :return string wrapper_version:
            Refprop wrapper version
        :return string refprop_version:
            Version of used RefProp dll
        """
        return self.__version__, self.rp.RPVersion()

    def is_mixture(self):
        """ Find out if fluid is mixture or not.
        In case current fluid is mixture, true is returned.
        In case current fluid is pure substance, false is returned.

        Return:
        -------
        :return boolean _mix_flag:
            Boolean for mixture (True), pure substance (False)
        """
        return self._mix_flag

    def set_error_flag(self, flag):
        """ Set error flag

        Parameters:
        :param boolean flag:
            New error flag
        Return:
        :return int err:
            Notifier for error code - in case everything went fine, 0 is returned
        """
        self._flag_check_errors = flag
        return 0

    def set_warning_flag(self, flag):
        """ Set warning flag

        Parameters:
        :param boolean flag:
            New warning flag
        Return:
        :return int err:
            Notifier for error code - in case everything went fine, 0 is returned
        """
        self._flag_warnings = flag
        return 0



"""Setup.py script for the vclibpy-framework"""

import setuptools

# read the contents of your README file
from pathlib import Path
readme_path = Path(__file__).parent.joinpath("README.md")
long_description = readme_path.read_text()

INSTALL_REQUIRES = [
    'numpy',
    'coolprop',
    'matplotlib',
    'pandas',
    'sdf',
    'ctREFPROP'
]

with open(Path(__file__).parent.joinpath("vclibpy", "__init__.py"), "r") as file:
    for line in file.readlines():
        if line.startswith("__version__"):
            VERSION = line.replace("__version__", "").split("=")[1].strip().replace("'", "").replace('"', '')

setuptools.setup(
    name='vclibpy',
    version=VERSION,
    description="Package with functions for the vapor "
                "compression process design using steady state simulation models",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/RWTH-EBC/vclibpy',
    download_url=f'https://github.com/RWTH-EBC/vclibpy/archive/refs/tags/{VERSION}.tar.gz',
    license='BSD 3-Clause',
    author='RWTH Aachen University, E.ON Energy Research Center, Institute '
           'of Energy Efficient Buildings and Indoor Climate',
    author_email='fabian.wuellhorst@eonerc.rwth-aachen.de',
    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    keywords=[
        "vapor-compression", "heat pumps",
        "chillers", "steady-state simulation"
    ],
    packages=setuptools.find_packages(exclude=['tests', 'tests.*', 'img']),
    install_requires=INSTALL_REQUIRES,
)



"""
Module with cool prop wrapper.
"""
import logging

import CoolProp.CoolProp as CoolPropInternal

from vclibpy.media.media import MedProp
from vclibpy.media.states import ThermodynamicState, TransportProperties

logger = logging.getLogger(__name__)


class CoolProp(MedProp):
    """
    Class using the open-source CoolProp package
    to access the properties.

    Args:
        use_high_level_api (bool):
            True to use the high-level api, which is much slower,
            but you can use all modes in calc_state.
            Default is False.
    """

    _mode_map = {
        "PT": (CoolPropInternal.PT_INPUTS, True),
        "TQ": (CoolPropInternal.QT_INPUTS, False),
        "PS": (CoolPropInternal.PSmass_INPUTS, True),
        "PH": (CoolPropInternal.HmassP_INPUTS, False),
        "PQ": (CoolPropInternal.PQ_INPUTS, True)
    }

    def __init__(self, fluid_name, use_high_level_api: bool = False):
        super().__init__(fluid_name=fluid_name)
        # Set molar mass and trigger a possible fluid-name error
        # if the fluid is not supported.
        self._helmholtz_equation_of_state = CoolPropInternal.AbstractState("HEOS", self.fluid_name)
        self.M = self._helmholtz_equation_of_state.molar_mass()
        self.use_high_level_api = use_high_level_api

    def calc_state(self, mode: str, var1: float, var2: float):
        super().calc_state(mode=mode, var1=var1, var2=var2)

        if self.use_high_level_api:
            _var1_str, _var2_str = mode[0], mode[1]
            # CoolProp returns Pa
            p = CoolPropInternal.PropsSI('P', _var1_str, var1, _var2_str, var2, self.fluid_name)
            # CoolProp returns kg/m^3
            d = CoolPropInternal.PropsSI('D', _var1_str, var1, _var2_str, var2, self.fluid_name)
            # CoolProp returns K
            T = CoolPropInternal.PropsSI('T', _var1_str, var1, _var2_str, var2, self.fluid_name)
            # CoolProp returns J/kg
            u = CoolPropInternal.PropsSI('U', _var1_str, var1, _var2_str, var2, self.fluid_name)
            # CoolProp returns J/kg
            h = CoolPropInternal.PropsSI('H', _var1_str, var1, _var2_str, var2, self.fluid_name)
            # CoolProp returns J/kg/K
            s = CoolPropInternal.PropsSI('S', _var1_str, var1, _var2_str, var2, self.fluid_name)
            # CoolProp returns mol/mol
            q = CoolPropInternal.PropsSI('Q', _var1_str, var1, _var2_str, var2, self.fluid_name)
            # Return new state
            return ThermodynamicState(p=p, T=T, u=u, h=h, s=s, d=d, q=q)

        self._update_coolprop_heos(mode=mode, var1=var1, var2=var2)
        # Return new state
        return ThermodynamicState(
            p=self._helmholtz_equation_of_state.p(),
            T=self._helmholtz_equation_of_state.T(),
            u=self._helmholtz_equation_of_state.umass(),
            h=self._helmholtz_equation_of_state.hmass(),
            s=self._helmholtz_equation_of_state.smass(),
            d=self._helmholtz_equation_of_state.rhomass(),
            q=self._helmholtz_equation_of_state.Q()
        )

    def calc_transport_properties(self, state: ThermodynamicState):
        if 0 <= state.q <= 1:
            mode = "PQ"
            var1, var2 = state.p, state.q
        else:
            # Get using p and T
            mode = "PT"
            var1, var2 = state.p, state.T

        if self.use_high_level_api:
            args = [mode[0], var1, mode[1], var2, self.fluid_name]
            # CoolProp returns -
            pr = CoolPropInternal.PropsSI('PRANDTL', *args)
            # CoolProp returns J/kg/K
            cp = CoolPropInternal.PropsSI('C', *args)
            # CoolProp returns J/kg/K
            cv = CoolPropInternal.PropsSI('CVMASS', *args)
            # CoolProp returns W/m/K
            lam = CoolPropInternal.PropsSI('CONDUCTIVITY', *args)
            # CoolProp returns Pa*s
            dyn_vis = CoolPropInternal.PropsSI('VISCOSITY', *args)
            # Internal calculation as kinematic vis is ration of dyn_vis to density
            # In m^2/s
            kin_vis = dyn_vis / state.d

            # Create transport properties instance
            return TransportProperties(lam=lam, dyn_vis=dyn_vis, kin_vis=kin_vis,
                                       pr=pr, cp=cp, cv=cv, state=state)
        # Low-level API
        self._update_coolprop_heos(mode=mode, var1=var1, var2=var2)
        # Create transport properties instance
        return TransportProperties(
            lam=self._helmholtz_equation_of_state.conductivity(),
            dyn_vis=self._helmholtz_equation_of_state.viscosity(),
            kin_vis=self._helmholtz_equation_of_state.viscosity() / state.d,
            pr=self._helmholtz_equation_of_state.Prandtl(),
            cp=self._helmholtz_equation_of_state.cpmass(),
            cv=self._helmholtz_equation_of_state.cvmass(),
            state=state
        )

    def _update_coolprop_heos(self, mode, var1, var2):
        if mode not in self._mode_map:
            raise KeyError(
                f"Given mode '{mode}' is currently not supported with the "
                f"faster low-level API to cool-prop. "
                f"Either use the high-level API or raise an issue. "
                f"Supported modes: {', '.join(self._mode_map.keys())}"
            )
        i_input, not_reverse_variables = self._mode_map[mode]
        if not_reverse_variables:
            self._helmholtz_equation_of_state.update(i_input, var1, var2)
        else:
            self._helmholtz_equation_of_state.update(i_input, var2, var1)

    def get_critical_point(self):
        Tc = CoolPropInternal.PropsSI("TCRIT", self.fluid_name)
        pc = CoolPropInternal.PropsSI("PCRIT", self.fluid_name)
        dc = CoolPropInternal.PropsSI("RHOCRIT", self.fluid_name)
        return Tc, pc, dc

    def get_molar_mass(self):
        return self.M


if __name__ == '__main__':
    CoolProp("Propan")



import pathlib
from typing import List

import numpy as np
import matplotlib.pyplot as plt
import sdf

from vclibpy.media import ThermodynamicState, MedProp


def plot_cycle(
        med_prop: MedProp,
        states: List[ThermodynamicState],
        save_path: pathlib.Path = None,
        show: bool = False
):
    """
    Creates T-h and p-h diagrams for a thermodynamic cycle.

    Args:
        med_prop (MedProp):
            Object containing the medium properties and two-phase limits.
        states (List[ThermodynamicState]):
            List of thermodynamic states defining the cycle points. Each state
            should contain T, p, and h properties.
        save_path (pathlib.Path, optional):
            Path where the plot should be saved. If None, returns the figure
            and axes objects instead. Defaults to None.
        show (bool):
            If True, plots are displayed. Default is False.

    Returns:
        tuple(matplotlib.figure.Figure, numpy.ndarray) or None:
            If save_path is provided, saves the plot and returns None.
            If show is True, shows the plot.
    """
    states.append(states[0])  # Plot full cycle
    # Unpack state var:
    h_T = np.array([state.h for state in states]) / 1000
    T = [state.T - 273.15 for state in states]
    p = np.array([state.p / 1e5 for state in states])
    h_p = h_T

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].set_ylabel("$T$ in °C")
    ax[1].set_xlabel("$h$ in kJ/kgK")
    # Two phase limits
    ax[0].plot(
        med_prop.get_two_phase_limits("h") / 1000,
        med_prop.get_two_phase_limits("T") - 273.15, color="black"
    )

    ax[0].plot(h_T, T, color="r", marker="s")
    ax[1].set_yscale('log')  # Set y-axis to logarithmic scale
    ax[1].plot(h_p, p, marker="s", color="r")
    # Two phase limits
    ax[1].plot(
        med_prop.get_two_phase_limits("h") / 1000,
        med_prop.get_two_phase_limits("p") / 1e5,
        color="black"
    )
    ax[1].set_ylabel("$log(p)$ in bar")
    ax[1].set_ylim([np.min(p) * 0.9, np.max(p) * 1.1])
    ax[0].set_ylim([np.min(T) - 5, np.max(T) + 5])
    ax[1].set_xlim([np.min(h_T) * 0.9, np.max(h_T) * 1.1])
    ax[0].set_xlim([np.min(h_T) * 0.9, np.max(h_T) * 1.1])
    if show:
        plt.show()
    if save_path is not None:
        fig.tight_layout()
        fig.savefig(save_path)
        plt.close(fig)
        return
    return fig, ax


def plot_sdf_map(
        filepath_sdf: pathlib.Path,
        nd_data: str,
        first_dimension: str,
        second_dimension: str,
        fluids: List[str] = None,
        flowsheets: List[str] = None,
        violin_plot_variable: str = None,
        third_dimension: str = None
):
    """
    Generate and display visualizations based on data from an SDF (Structured Data File) dataset.
    This function generates various types of visualizations based on the provided parameters,
    including 3D scatter plots, 3D surface plots, and violin plots, and displays them using Matplotlib.

    Args:
        filepath_sdf (pathlib.Path):
            The path to the SDF dataset file.
        nd_data (str):
            The name of the primary data to be plotted.
        first_dimension (str):
            The name of the first dimension for the visualization.
        second_dimension (str):
            The name of the second dimension for the visualization.
        fluids (List[str], optional):
            List of specific fluids to include in the visualization.
            Default is None, which includes all fluids.
        flowsheets (List[str], optional):
            List of specific flowsheets to include in the visualization.
            Default is None, which includes all flowsheets.
        violin_plot_variable (str, optional):
            The variable to be used for creating violin plots.
            Default is None, which disables violin plots.
        third_dimension (str, optional):
            The name of the third dimension for 4D visualizations.
            Default is None, which disables 4D plotting.

    Raises:
        KeyError: If the specified data or dimensions are not found in the dataset.

    Examples:
    >>> FILEPATH_SDF = r"HeatPumpMaps.sdf"
    >>> plot_sdf_map(
    >>>     filepath_sdf=FILEPATH_SDF,
    >>>     nd_data="COP",
    >>>     first_dimension="T_eva_in",
    >>>     second_dimension="n",
    >>>     fluids=["R410A"],
    >>>     flowsheets=["OptiHorn"],
    >>> )

    """
    if fluids is None:
        fluids = []
    if flowsheets is None:
        flowsheets = []
    if "T_" in second_dimension:
        offset_sec = -273.15
    else:
        offset_sec = 0
    if "T_" in first_dimension:
        offset_pri = -273.15
    else:
        offset_pri = 0
    offset_thi = 0
    plot_4D = False
    if third_dimension is not None:
        plot_4D = True
        if "T_" in third_dimension:
            offset_thi = -273.15

    dataset = sdf.load(str(filepath_sdf))
    plot_violin = True
    if violin_plot_variable is None:
        plot_violin = False
        violin_plot_variable = ""
    if plot_violin:
        if flowsheets:
            n_rows = len(flowsheets)
        else:
            n_rows = len(dataset.groups)
        fig_v, ax_v = plt.subplots(nrows=n_rows, ncols=1, sharex=True,
                                   squeeze=False)
        fig_v.suptitle(violin_plot_variable)
    i_fs = 0
    nd_str_plot = nd_data
    fac = 1
    if nd_data == "dT_eva_min":
        nd_data = "T_1"
        sub_str = "T_eva_in"
        fac = - 1
    elif nd_data == "dT_con":
        nd_data = "T_3"
        sub_str = "T_con_in"
    elif nd_data == "dT_sh":
        nd_data = "T_1"
        sub_str = "T_4"
    else:
        sub_str = ""

    if nd_str_plot.startswith("T_"):
        offset_nd = -273.15
    else:
        offset_nd = 0

    for flowsheet in dataset.groups:
        violin_data = {}
        if flowsheet.name not in flowsheets and len(flowsheets) > 0:
            continue
        for fluid in flowsheet.groups:
            if fluid.name not in fluids and len(fluids) > 0:
                continue
            nd, fd, sd, sub_data, td = None, None, None, None, None
            _other_scale = {}
            for ds in fluid.datasets:
                if ds.name == nd_data:
                    nd = ds
                elif ds.name == first_dimension:
                    fd = ds.data
                elif ds.name == second_dimension:
                    sd = ds.data
                if ds.name == sub_str:
                    sub_data = ds.data
                if ds.name == violin_plot_variable:
                    data = ds.data.flatten()
                    violin_data[fluid.name] = data[~np.isnan(data)]
                if plot_4D and ds.name == third_dimension:
                    td = ds.data

            if nd is None:
                raise KeyError("nd-String not found in dataset")

            if sub_data is None:
                sub_data = np.zeros(nd.data.shape)

            # Check other scales:
            for i, scale in enumerate(nd.scales):
                if scale.name not in [first_dimension, second_dimension]:
                    _other_scale[i] = scale

            if fd is None or sd is None or not _other_scale or (plot_4D and td is None):
                raise KeyError("One of the given strings was not found in dataset")

            if plot_4D:
                fig = plt.figure()
                figtitle = f"{flowsheet.name}_{fluid.name}_{nd_str_plot}"
                fig.suptitle(figtitle)
                ax = fig.add_subplot(111, projection='3d')
                ax.set_xlabel(first_dimension)
                ax.set_ylabel(second_dimension)
                ax.set_zlabel(third_dimension)
                fourth_dim = (nd.data - sub_data) * fac + offset_nd
                # Scale values for better sizes of circles:
                bounds = [fourth_dim.min(), fourth_dim.max()]
                _max_circle_size = 30
                fourth_dim_scaled = (fourth_dim - bounds[0]) / (bounds[1] - bounds[0]) * _max_circle_size
                inp = [fd + offset_pri, sd + offset_sec, td + offset_thi]
                import itertools
                scattergrid = np.array([c for c in itertools.product(*inp)])
                ax.scatter(scattergrid[:, 0],
                           scattergrid[:, 1],
                           scattergrid[:, 2],
                           c=fourth_dim_scaled,
                           s=fourth_dim_scaled)
            else:
                for index, scale in _other_scale.items():
                    for idx_data, value in enumerate(scale.data):
                        if index==0:
                            Z = nd.data[idx_data, :, :]
                            if sub_str in ["T_4", ""]:
                                sub_data_use = sub_data[idx_data, :, :]
                            else:
                                sub_data_use = sub_data
                        elif index == 1:
                            Z = nd.data[:, idx_data, :]
                            if sub_str in ["T_4", ""]:
                                sub_data_use = sub_data[:, idx_data, :]
                            else:
                                sub_data_use = sub_data
                        else:
                            Z = nd.data[:, :, idx_data]
                            if sub_str in ["T_4", ""]:
                                sub_data_use = sub_data[:, :, idx_data]
                            else:
                                sub_data_use = sub_data
                        if not plot_violin:
                            fig = plt.figure()
                            figtitle = f"{flowsheet.name}_{fluid.name}_{nd_str_plot}_{scale.name}={round(value, 3)}"
                            fig.suptitle(figtitle)
                            ax = fig.add_subplot(111, projection='3d')
                            ax.set_xlabel(first_dimension)
                            ax.set_ylabel(second_dimension)
                            X, Y = np.meshgrid(fd, sd)
                            ax.plot_surface(X + offset_pri, Y + offset_sec, (Z - sub_data_use)*fac + offset_nd)

        if plot_violin:
            for key, value in violin_data.items():
                print(f"{violin_plot_variable}: {flowsheet.name}_{key}")
                print(f"Min: {np.min(value)}")
                print(f"Max: {np.max(value)}")
                print(f"Mean: {np.mean(value)}")
                print(f"Median: {np.median(value)}\n")
            ax_v[i_fs][0].violinplot(
                                     list(violin_data.values()),
                                     showextrema=True,
                                     showmeans=True,
                                     showmedians=True
                                     )
            set_axis_style(ax_v[i_fs][0], list(violin_data.keys()))
            ax_v[i_fs][0].set_ylabel(flowsheet.name.replace("Flowsheet", ""))
        i_fs += 1
    plt.show()


def set_axis_style(ax, labels):
    """
    From: https://matplotlib.org/3.1.1/gallery/statistics/customized_violin.html#sphx-glr-gallery-statistics-customized-violin-py
    """
    ax.get_xaxis().set_tick_params(direction='out')
    ax.xaxis.set_ticks_position('bottom')
    ax.set_xticks(np.arange(1, len(labels) + 1))
    ax.set_xticklabels(labels)
    ax.set_xlim(0.25, len(labels) + 0.75)



import abc
import logging

from .heat_transfer import HeatTransfer
from vclibpy.media import TransportProperties


logger = logging.getLogger(__name__)


class AirToWallTransfer(HeatTransfer, abc.ABC):
    """
    Heat transfer model for air to wall.

    Args:
        A_cross (float):
            Cross-section area in m2.
        characteristic_length (float):
            Length in m to calculate the similitude approach for the
            heat transfer from secondary_medium -> wall.
    """

    def __init__(self, A_cross: float, characteristic_length: float):
        self.A_cross = A_cross
        self.characteristic_length = characteristic_length

    def calc(self, transport_properties: TransportProperties, m_flow: float) -> float:
        """
        Heat transfer coefficient from air to the wall of the heat exchanger.
        The flow is assumed to be always laminar.

        Returns:
            float: Heat transfer coefficient in W/(m^2*K).
        """
        Re = self.calc_reynolds(dynamic_viscosity=transport_properties.dyn_vis, m_flow=m_flow)
        alpha_sec = self.calc_laminar_area_nusselt(Re, transport_properties.Pr, lambda_=transport_properties.lam)
        return alpha_sec

    @abc.abstractmethod
    def calc_reynolds(self, dynamic_viscosity: float, m_flow: float):
        """
        Calculate the reynolds number of the given flow rate.

        Args:
            dynamic_viscosity (float): Dynamic viscosity.
            m_flow (float): Mass flow rate.

        Returns:
            float: Reynolds number
        """
        raise NotImplementedError

    @abc.abstractmethod
    def calc_laminar_area_nusselt(self, Re, Pr, lambda_):
        """
        Calculate the Nusselt number for laminar heat transfer
        on an area (Used for Air->Wall in the evaporator).

        Args:
            Re (float): Reynolds number
            Pr (float): Prandlt number
            lambda_ (float): Lambda of air

        Returns:
            float: Nusselt number
        """
        raise NotImplementedError


class WSUAirToWall(AirToWallTransfer):
    """
    Class to implement the heat transfer calculations
    based on the WSÜ-Script at the RWTH.
    """

    def calc_reynolds(self, dynamic_viscosity: float, m_flow: float) -> float:
        """
        Calculate the Reynolds number on air side.

        Args:
            dynamic_viscosity (float): Dynamic viscosity of air.
            m_flow (float): Mass flow rate of air.

        Returns:
            float: Reynolds number.
        """
        velocity_times_dens = m_flow / self.A_cross
        return (velocity_times_dens * self.characteristic_length) / dynamic_viscosity

    def calc_laminar_area_nusselt(self, Re, Pr, lambda_) -> float:
        """
        Calculate the Nusselt number for laminar heat transfer
        on an area (Used for Air->Wall in the evaporator).

        Args:
            Re (float): Reynolds number of air.
            Pr (float): Prandtl number of air.
            lambda_ (float): Lambda of air.

        Returns:
            float: Nusselt number of air.
        """
        const_fac = 0.664
        exp_rey = 0.5
        exp_pra = 1 / 3
        if Re > 2e5:
            raise ValueError(f"Given Re {Re} is outside of allowed bounds for WSÜ-Script")
        if Pr > 10 or Pr < 0.6:
            raise ValueError(f"Given Pr {Pr} is outside of allowed bounds for WSÜ-Script")
        Nu = const_fac * Re ** exp_rey * Pr ** exp_pra
        return Nu * lambda_ / self.characteristic_length



import logging
from typing import List
import numpy as np

from abc import abstractmethod
import matplotlib.pyplot as plt
from vclibpy import media, Inputs
from vclibpy.datamodels import FlowsheetState
from vclibpy.components.heat_exchangers import HeatExchanger
from vclibpy.components.component import BaseComponent
from vclibpy.utils.plotting import plot_cycle

logger = logging.getLogger(__name__)


class BaseCycle:
    """
    Base class for a heat pump. More complex systems may inherit from this class
    All HP have a compressor, two HE and a source and sink.
    Therefore, the parameters defined here are general parameters.

    Args:
        fluid (str): Name of the fluid
        evaporator (HeatExchanger): Instance of a heat exchanger used for the  evaporator
        condenser (HeatExchanger): Instance of a heat exchanger used for the condenser
     """

    flowsheet_name: str = "BaseCLass of all HP classes - not to use for map generation"

    def __init__(
            self,
            fluid: str,
            evaporator: HeatExchanger,
            condenser: HeatExchanger
    ):
        self.fluid: str = fluid
        self.evaporator = evaporator
        self.condenser = condenser
        # Instantiate dummy values
        self.med_prop = None
        self._p_min = 10000  # So that p>0 at all times
        self._p_max = None  # Is set by med-prop

    def __str__(self):
        return self.flowsheet_name

    def setup_new_fluid(self, fluid):
        # Only do so if new fluid is given
        if self.med_prop is not None:
            if self.med_prop.fluid_name == fluid:
                return
            self.med_prop.terminate()

        # Else create new instance of MedProp
        med_prop_class, med_prop_kwargs = media.get_global_med_prop_and_kwargs()
        self.med_prop = med_prop_class(fluid_name=fluid, **med_prop_kwargs)

        # Write the instance to the components
        for component in self.get_all_components():
            component.med_prop = self.med_prop
            component.start_secondary_med_prop()

        # Get max and min pressure
        _, self._p_max, _ = self.med_prop.get_critical_point()
        self.fluid = fluid

    def terminate(self):
        self.med_prop.terminate()
        for component in self.get_all_components():
            component.terminate_secondary_med_prop()

    def get_all_components(self) -> List[BaseComponent]:
        return [self.condenser, self.evaporator]

    def calc_steady_state(self, inputs: Inputs, fluid: str = None, **kwargs):
        """
        Calculate the steady-state performance of a vapor compression cycle
        based on given inputs and assumptions.

        This function ensures consistent assumptions across different cycles.
        It calculates the performance of the heat pump under
        specific conditions while adhering to several general assumptions.

        General Assumptions:
        ---------------------
        - Isenthalpic expansion valves:
          The enthalpy at the inlet equals the enthalpy at the outlet.
        - No heat losses in any component:
          The heat input to the condenser equals the heat
          output of the evaporator plus the power input.
        - Input to the evaporator is always in the two-phase region.
        - Output of the evaporator and output of the condenser maintain
          a constant overheating or subcooling (can be set in Inputs).

        Args:
            inputs (Inputs):
                An instance of the Inputs class containing the
                necessary parameters to calculate the flowsheet state.
            fluid (str):
                The fluid to be used in the calculations.
                Required only if 'fluid' is not specified during the object's initialization.

        Keyword Arguments:
            min_iteration_step (int):
                The minimum step size for iterations (default: 1).
            save_path_plots (str or None):
                The path to save plots (default: None).
                If None, no plots are created.
            show_iteration (bool):
                Whether to display iteration progress (default: False).
            T_max (float):
                Maximum temperature allowed (default: 273.15 + 150).
            use_quick_solver (bool):
                Whether to use a quick solver (default: True).
            max_err_ntu (float):
                Maximum allowable error for the heat exchanger in percent (default: 0.5).
            max_err_dT_min (float):
                Maximum allowable error for minimum temperature difference in K (default: 0.1).
            max_num_iterations (int or None):
                Maximum number of iterations allowed (default: None).

        Returns:
            fs_state (FlowsheetState):
                An instance of the FlowsheetState class representing
                the calculated state of the vapor compression cycle.
        """
        # Settings
        min_iteration_step = kwargs.pop("min_iteration_step", 1)
        save_path_plots = kwargs.get("save_path_plots", None)
        input_name = ";".join([k + "=" + str(np.round(v.value, 3)).replace(".", "_")
                               for k, v in inputs.get_variables().items()])
        show_iteration = kwargs.get("show_iteration", False)
        use_quick_solver = kwargs.pop("use_quick_solver", True)
        err_ntu = kwargs.pop("max_err_ntu", 0.5)
        err_dT_min = kwargs.pop("max_err_dT_min", 0.1)
        max_num_iterations = kwargs.pop("max_num_iterations", 1e5)
        p_1_history = []
        p_2_history = []

        if use_quick_solver:
            step_p1 = kwargs.get("step_max", 10000)
            step_p2 = kwargs.get("step_max", 10000)
        else:
            step_p1 = min_iteration_step
            step_p2 = min_iteration_step

        # Setup fluid:
        if fluid is None:
            fluid = self.fluid
        self.setup_new_fluid(fluid)

        # First: Iterate with given conditions to get the 4 states and the mass flow rate:
        T_1_start = inputs.T_eva_in - inputs.dT_eva_superheating
        T_3_start = inputs.T_con_in + inputs.dT_con_subcooling
        p_1_start = self.med_prop.calc_state("TQ", T_1_start, 1).p
        p_2_start = self.med_prop.calc_state("TQ", T_3_start, 0).p
        p_1_next = p_1_start
        p_2_next = p_2_start

        fs_state = FlowsheetState()  # Always log what is happening in the whole flowsheet
        fs_state.set(name="Q_con", value=1, unit="W", description="Condenser heat flow rate")
        fs_state.set(name="COP", value=0, unit="-", description="Coefficient of performance")

        if show_iteration:
            fig_iterations, ax_iterations = plt.subplots(2)

        num_iterations = 0

        while True:
            if isinstance(max_num_iterations, (int, float)):
                if num_iterations > max_num_iterations:
                    logger.warning("Maximum number of iterations %s exceeded. Stopping.",
                                   max_num_iterations)
                    return

                if (num_iterations + 1) % (0.1 * max_num_iterations) == 0:
                    logger.info("Info: %s percent of max_num_iterations %s used",
                                100 * (num_iterations + 1) / max_num_iterations, max_num_iterations)

            p_1 = p_1_next
            p_2 = p_2_next
            p_1_history.append(p_1)
            p_2_history.append(p_2)
            if show_iteration:
                ax_iterations[0].cla()
                ax_iterations[1].cla()
                ax_iterations[0].scatter(list(range(len(p_1_history))), p_1_history)
                ax_iterations[1].scatter(list(range(len(p_2_history))), p_2_history)
                plt.draw()
                plt.pause(1e-5)

            # Increase counter
            num_iterations += 1
            # Check critical pressures:
            if p_2 >= self._p_max:
                if step_p2 == min_iteration_step:
                    logger.error("Pressure too high. Configuration is infeasible.")
                    return
                p_2_next = p_2 - step_p2
                step_p2 /= 10
                continue
            if p_1 <= self._p_min:
                if p_1_next == min_iteration_step:
                    logger.error("Pressure too low. Configuration is infeasible.")
                    return
                p_1_next = p_1 + step_p1
                step_p1 /= 10
                continue

            # Calculate the states based on the given flowsheet
            try:
                self.calc_states(p_1, p_2, inputs=inputs, fs_state=fs_state)
            except ValueError as err:
                logger.error("An error occurred while calculating states. "
                             "Can't guess next pressures, thus, exiting: %s", err)
                return
            if save_path_plots is not None and num_iterations == 1 and show_iteration:
                self.plot_cycle(save_path=save_path_plots.joinpath(f"{input_name}_initialization.png"), inputs=inputs)

            # Check heat exchangers:
            error_eva, dT_min_eva = self.evaporator.calc(inputs=inputs, fs_state=fs_state)
            if not isinstance(error_eva, float):
                print(error_eva)
            if error_eva < 0:
                p_1_next = p_1 - step_p1
                continue
            else:
                if step_p1 > min_iteration_step:
                    p_1_next = p_1 + step_p1
                    step_p1 /= 10
                    continue
                elif error_eva > err_ntu and dT_min_eva > err_dT_min:
                    step_p1 = 1000
                    p_1_next = p_1 + step_p1
                    continue

            error_con, dT_min_con = self.condenser.calc(inputs=inputs, fs_state=fs_state)
            if error_con < 0:
                p_2_next = p_2 + step_p2
                continue
            else:
                if step_p2 > min_iteration_step:
                    p_2_next = p_2 - step_p2
                    step_p2 /= 10
                    continue
                elif error_con > err_ntu and dT_min_con > err_dT_min:
                    p_2_next = p_2 - step_p2
                    step_p2 = 1000
                    continue

            # If still here, and the values are equal, we may break.
            if p_1 == p_1_next and p_2 == p_2_next:
                # Check if solution was too far away. If so, jump back
                # And decrease the iteration step by factor 10.
                if step_p2 > min_iteration_step:
                    p_2_next = p_2 - step_p2
                    step_p2 /= 10
                    continue
                if step_p1 > min_iteration_step:
                    p_1_next = p_1 + step_p1
                    step_p1 /= 10
                    continue
                logger.info("Breaking: Converged")
                break

            # Check if values are not converging at all:
            p_1_unique = set(p_1_history[-10:])
            p_2_unique = set(p_2_history[-10:])
            if len(p_1_unique) == 2 and len(p_2_unique) == 2 \
                    and step_p1 == min_iteration_step and step_p2 == min_iteration_step:
                logger.critical("Breaking: not converging at all")
                break

        if show_iteration:
            plt.close(fig_iterations)

        # Calculate the heat flow rates for the selected states.
        Q_con = self.condenser.calc_Q_flow()
        Q_con_outer = self.condenser.calc_secondary_Q_flow(Q_con)
        Q_eva = self.evaporator.calc_Q_flow()
        Q_eva_outer = self.evaporator.calc_secondary_Q_flow(Q_eva)
        self.evaporator.calc(inputs=inputs, fs_state=fs_state)
        self.condenser.calc(inputs=inputs, fs_state=fs_state)
        P_el = self.calc_electrical_power(fs_state=fs_state, inputs=inputs)
        T_con_out = inputs.T_con_in + Q_con_outer / self.condenser.m_flow_secondary_cp

        # COP based on P_el and Q_con:
        COP_inner = Q_con / P_el
        COP_outer = Q_con_outer / P_el
        # Calculate carnot quality as a measure of reliability of model:
        COP_carnot = (T_con_out / (T_con_out - inputs.T_eva_in))
        carnot_quality = COP_inner / COP_carnot
        # Calc return temperature:
        fs_state.set(
            name="P_el", value=P_el, unit="W",
            description="Power consumption"
        )
        fs_state.set(
            name="carnot_quality", value=carnot_quality,
            unit="-", description="Carnot Quality"
        )
        fs_state.set(
            name="Q_con", value=Q_con, unit="W",
            description="Condenser refrigerant heat flow rate"
        )
        # COP based on P_el and Q_con:
        fs_state.set(
            name="Q_con_outer", value=Q_con_outer, unit="W",
            description="Secondary medium condenser heat flow rate"
        )
        fs_state.set(
            name="Q_eva_outer", value=Q_eva_outer, unit="W",
            description="Secondary medium evaporator heat flow rate"
        )
        fs_state.set(
            name="COP", value=COP_inner,
            unit="-", description="Coefficient of Performance"
        )
        fs_state.set(
            name="COP_outer", value=COP_outer,
            unit="-", description="Outer COP, including heat losses"
        )

        if save_path_plots is not None:
            self.plot_cycle(save_path=save_path_plots.joinpath(f"{input_name}_final_result.png"), inputs=inputs)

        return fs_state

    @abstractmethod
    def get_states_in_order_for_plotting(self):
        """
        Function to return all thermodynamic states of cycle
        in the correct order for plotting.
        Include phase change states to see if your simulation
        runs plausible cycles.

        Returns:
            - List with tuples, first entry being the state and second the mass flow rate
        """
        return []

    def set_evaporator_outlet_based_on_superheating(self, p_eva: float, inputs: Inputs):
        """
        Calculate the outlet state of the evaporator based on
        the required degree of superheating.

        Args:
            p_eva (float): Evaporation pressure
            inputs (Inputs): Inputs with superheating level
        """
        T_1 = self.med_prop.calc_state("PQ", p_eva, 1).T + inputs.dT_eva_superheating
        if inputs.dT_eva_superheating > 0:
            self.evaporator.state_outlet = self.med_prop.calc_state("PT", p_eva, T_1)
        else:
            self.evaporator.state_outlet = self.med_prop.calc_state("PQ", p_eva, 1)

    def set_condenser_outlet_based_on_subcooling(self, p_con: float, inputs: Inputs):
        """
        Calculate the outlet state of the evaporator based on
        the required degree of superheating.

        Args:
            p_con (float): Condensing pressure
            inputs (Inputs): Inputs with superheating level
        """
        T_3 = self.med_prop.calc_state("PQ", p_con, 0).T - inputs.dT_con_subcooling
        if inputs.dT_con_subcooling > 0:
            self.condenser.state_outlet = self.med_prop.calc_state("PT", p_con, T_3)
        else:
            self.condenser.state_outlet = self.med_prop.calc_state("PQ", p_con, 0)

    def plot_cycle(self, save_path: str, inputs: Inputs):
        """Function to plot the resulting flowsheet of the steady state config."""
        states = self.get_states_in_order_for_plotting()
        fig, ax = plot_cycle(
            states=states,
            med_prop=self.med_prop,
            save_path=None
        )
        self._plot_secondary_heat_flow_rates(ax=ax[0], inputs=inputs)
        fig.tight_layout()
        fig.savefig(save_path)
        plt.close(fig)

    def _plot_secondary_heat_flow_rates(self, ax, inputs):
        Q_con = self.condenser.calc_Q_flow()
        Q_eva = self.evaporator.calc_Q_flow()

        delta_H_con = np.array([
            self.condenser.state_outlet.h * self.condenser.m_flow,
            self.condenser.state_outlet.h * self.condenser.m_flow + Q_con
        ]) / self.condenser.m_flow
        delta_H_eva = np.array([
            self.evaporator.state_outlet.h * self.evaporator.m_flow,
            self.evaporator.state_outlet.h * self.evaporator.m_flow - Q_eva
        ]) / self.evaporator.m_flow
        self.condenser.m_flow_secondary = inputs.m_flow_con
        self.condenser.calc_secondary_cp(T=inputs.T_con_in)
        self.evaporator.m_flow_secondary = inputs.m_flow_eva
        self.evaporator.calc_secondary_cp(T=inputs.T_eva_in)
        ax.plot(delta_H_con / 1000, [
            inputs.T_con_in - 273.15,
            inputs.T_con_in + Q_con / self.condenser.m_flow_secondary_cp - 273.15
        ], color="b")
        ax.plot(delta_H_eva / 1000, [
            inputs.T_eva_in - 273.15,
            inputs.T_eva_in - Q_eva / self.evaporator.m_flow_secondary_cp - 273.15
        ], color="b")

    @abstractmethod
    def calc_electrical_power(self, inputs: Inputs, fs_state: FlowsheetState):
        """Function to calc the electrical power consumption based on the flowsheet used"""
        raise NotImplementedError

    @abstractmethod
    def calc_states(self, p_1, p_2, inputs: Inputs, fs_state: FlowsheetState):
        """
        Function to calculate the states and mass flow rates of the flowsheet
        and set these into each component based on the given pressure levels p_1 and p_2.

        Args:
            p_1 (float):
                Lower pressure level. If no pressure losses are assumed,
                this equals the evaporation pressure and the compressor inlet pressure.
            p_2 (float):
                Higher pressure level. If no pressure losses are assumed,
                this equals the condensing pressure and the compressor outlet pressure.
            inputs (Inputs): Inputs of calculation.
            fs_state (FlowsheetState): Flowsheet state to save important variables.
        """
        raise NotImplementedError
